HAVE_VTK = False
try:
    import vtk
    HAVE_VTK = True
    from vtk.util.numpy_support import vtk_to_numpy
except ImportError:
    pass


def vtk_image_to_ndarray(image):
    array = vtk_to_numpy(image.GetPointData().GetScalars())
    dims = list(image.GetDimensions())
    array.shape = dims[::-1]
    return array

def vtk_polydata_to_vtkjs(point_set):
    array = vtk_to_numpy(point_set.GetPoints().GetData())
    return array



from imjoy_rpc import api
from typing import List

from ._type_aliases import Gaussians, Style, Image, Point_Sets
from ._initialization_params import init_params_dict, init_key_aliases
from .integrations import _detect_render_type, _get_viewer_image, _get_viewer_point_sets
from .render_types import RenderType

__all__ = [
    "Viewer",
    "view",
]

_viewer_count = 1


class ViewerRPC:
    """Viewer remote procedure interface."""

    def __init__(
        self, ui_collapsed=True, rotate=False, ui="pydata-sphinx", **add_data_kwargs
    ):
        """Create a viewer."""
        self._init_viewer_kwargs = dict(ui_collapsed=ui_collapsed, rotate=rotate, ui=ui)
        self._init_viewer_kwargs.update(**add_data_kwargs)
        self.init_data = {}

    def _get_input_data(self):
        input_options = ["data", "image", "point_sets"]
        inputs = []
        for option in input_options:
            data = self._init_viewer_kwargs.get(option, None)
            if data is not None:
                inputs.append((option, data))
        return inputs

    async def setup(self):
        pass

    async def run(self, ctx):
        """ImJoy plugin setup function."""
        global _viewer_count
        ui = self._init_viewer_kwargs.get("ui", None)
        if ui == "pydata-sphinx":
            config = {
                "uiMachineOptions": {
                    "href": "https://cdn.jsdelivr.net/npm/itk-viewer-bootstrap-ui@0/dist/itk-viewer-bootstrap-ui.es.js",
                    "export": "default",
                }
            }
        elif ui == "mui":
            config = {
                "uiMachineOptions": {
                    "href": "https://www.jsdelivr.com/package/npm/itk-viewer-material-ui@0/dist/materialUIMachineOptions.js.es.js",
                    "export": "default",
                }
            }
        elif ui != "reference":
            config = ui
        else:
            config = {}

        inputs = self._get_input_data()

        self.init_data.clear()
        for (input_type, data) in inputs:
            render_type = _detect_render_type(data, input_type)
            key = init_key_aliases()[input_type]
            if render_type is RenderType.IMAGE:
                result = await _get_viewer_image(data)
            elif render_type is RenderType.POINT_SET:
                result = await _get_viewer_point_sets(data)
            if not result:
                result = data
            self.init_data[key] = result

        itk_viewer = await api.createWindow(
            name=f"itkwidgets viewer {_viewer_count}",
            type="itk-vtk-viewer",
            src="https://kitware.github.io/itk-vtk-viewer/app",
            fullscreen=False,
            data=self.init_data,
            # config should be a python data dictionary and can't be a string e.g. 'pydata-sphinx',
            config=config,
        )
        _viewer_count += 1

        self.set_default_ui_values(itk_viewer)
        self.itk_viewer = itk_viewer

    def set_default_ui_values(self, itk_viewer):
        settings = init_params_dict(itk_viewer)
        for key, value in self._init_viewer_kwargs.items():
            if key in settings.keys():
                settings[key](value)


class Viewer:
    """Pythonic Viewer class."""

    def __init__(
        self, ui_collapsed=True, rotate=False, ui="pydata-sphinx", **add_data_kwargs
    ):
        """Create a viewer."""
        self.viewer_rpc = ViewerRPC(
            ui_collapsed=ui_collapsed, rotate=rotate, ui=ui, **add_data_kwargs
        )
        api.export(self.viewer_rpc)

    def set_annotations_enabled(self, enabled: bool):
        self.viewer_rpc.itk_viewer.setAnnotationsEnabled(enabled)

    def set_axes_enabled(self, enabled: bool):
        self.viewer_rpc.itk_viewer.setAxesEnabled(enabled)

    def set_background_color(self, bgColor: List[float]):
        self.viewer_rpc.itk_viewer.setBackgroundColor(bgColor)

    async def set_image(self, image: Image):
        render_type = _detect_render_type(image, 'image')
        if render_type is RenderType.IMAGE:
            image = _get_viewer_image(image)
            await self.viewer_rpc.itk_viewer.setImage(image)
        elif render_type is RenderType.POINT_SET:
            image = _get_viewer_point_sets(image)
            await self.viewer_rpc.itk_viewer.setPointSets(image)

    def set_image_blend_mode(self, mode: str):
        self.viewer_rpc.itk_viewer.setImageBlendMode(mode)

    def set_image_color_map(self, colorMap: str):
        self.viewer_rpc.itk_viewer.setImageColorMap(colorMap)

    def set_image_color_range(self, range: List[float]):
        self.viewer_rpc.itk_viewer.setImageColorRange(range)

    def set_image_color_range_bounds(self, range: List[float]):
        self.viewer_rpc.itk_viewer.setImageColorRangeBounds(range)

    def set_image_component_visibility(self, visibility: bool):
        self.viewer_rpc.itk_viewer.setImageComponentVisibility(visibility)

    def set_image_gradient_opacity(self, opacity: float):
        self.viewer_rpc.itk_viewer.setImageGradientOpacity(opacity)

    def set_image_gradient_opacity_scale(self, min: float):
        self.viewer_rpc.itk_viewer.setImageGradientOpacityScale(min)

    def set_image_interpolation_enabled(self, enabled: bool):
        self.viewer_rpc.itk_viewer.setImageInterpolationEnabled(enabled)

    def set_image_piecewise_function_gaussians(self, gaussians: Gaussians):
        self.viewer_rpc.itk_viewer.setImagePiecewiseFunctionGaussians(gaussians)

    def set_image_shadow_enabled(self, enabled: bool):
        self.viewer_rpc.itk_viewer.setImageShadowEnabled(enabled)

    def set_image_volume_sample_distance(self, distance: float):
        self.viewer_rpc.itk_viewer.setImageVolumeSampleDistance(distance)

    async def set_label_image(self, label_image: Image):
        render_type = _detect_render_type(label_image, 'image')
        if render_type is RenderType.IMAGE:
            label_image = _get_viewer_image(label_image, is_label=True)
            await self.viewer_rpc.itk_viewer.setImage(label_image)
        elif render_type is RenderType.POINT_SET:
            label_image = _get_viewer_point_sets(label_image, is_label=True)
            await self.viewer_rpc.itk_viewer.setPointSets(label_image)

    def set_label_image_blend(self, blend: float):
        self.viewer_rpc.itk_viewer.setLabelImageBlend(blend)

    def set_label_image_label_names(self, names: List[str]):
        self.viewer_rpc.itk_viewer.setLabelImageLabelNames(names)

    def set_label_image_lookup_table(self, lookupTable: str):
        self.viewer_rpc.itk_viewer.setLabelImageLookupTable(lookupTable)

    def set_label_image_weights(self, weights: float):
        self.viewer_rpc.itk_viewer.setLabelImageWeights(weights)

    def select_layer(self, name: str):
        self.viewer_rpc.itk_viewer.selectLayer(name)

    def set_layer_visibility(self, visible: bool):
        self.viewer_rpc.itk_viewer.setLayerVisibility(visible)

    def set_point_sets(self, pointSets: Point_Sets):
        self.viewer_rpc.itk_viewer.setPointSets(pointSets)

    def set_rendering_view_container_style(self, containerStyle: Style):
        self.viewer_rpc.itk_viewer.setRenderingViewContainerStyle(containerStyle)

    def set_rotate(self, enabled: bool):
        self.viewer_rpc.itk_viewer.setRotateEnabled(enabled)

    def set_ui_collapsed(self, collapsed: bool):
        self.viewer_rpc.itk_viewer.setUICollapsed(collapsed)

    def set_units(self, units: str):
        self.viewer_rpc.itk_viewer.setUnits(units)

    def set_view_mode(self, mode: str):
        self.viewer_rpc.itk_viewer.setViewMode(mode)

    def set_x_slice(self, position: float):
        self.viewer_rpc.itk_viewer.setXSlice(position)

    def set_y_slice(self, position: float):
        self.viewer_rpc.itk_viewer.setYSlice(position)

    def set_z_slice(self, position: float):
        self.viewer_rpc.itk_viewer.setZSlice(position)


def view(data=None, **kwargs):
    """View the image and/or point sets.

    Creates and returns an ImJoy plugin ipywidget to visualize an image, and/or
    point sets.

    The image can be 2D or 3D. The type of the image can be an numpy.array,
    itk.Image, or vtk.vtkImageData.

    A point set or a sequence of points sets can be visualized. The type of the
    point set can be an numpy.array (Nx3 array of point positions).

    Parameters
    ----------
    General Interface
    ^^^^^^^^^^^^^^^^^
    ui_collapsed : bool, default: True
        Collapse the native widget user interface.
    rotate : bool, default: False
        Continuously rotate the camera around the scene in volume rendering
        mode.
    annotations: bool, default: True
        Display annotations describing orientation and the value of a
        mouse-position-based data probe.
    axes: bool, default: False
        Display axes.
    bg_color: (red, green, blue) tuple, components from 0.0 to 1.0
        Background color. Default is based on the current Jupyter theme.
    container_style: dict
        The CSS style for the rendering view `div`'s.

    Images
    ^^^^^^
    image : array_like, itk.Image, or vtk.vtkImageData
        The 2D or 3D image to visualize.
    label_blend: float, default: 0.5
        Label map blend with intensity image, from 0.0 to 1.0.
    label_names: list of (label_value, label_name)
        String names associated with the integer label values.
    label_lut: string, default: 'glasbey'
        Lookup table for the label map.
    label_weights: float
        The rendering weight assigned to current label. Values range from 0.0
        to 1.0.
    color_range: list, default: The [min, max] range of the data values
        The [min, max] range of intensity values mapped to colors for the given
        image component identified by name.
    color_bounds: list, default: The [min, max] range of the data values
        The [min, max] range of intensity values for color maps that provide a
        bounds for user inputs.
    cmap: string, default: 'Grayscale'
        The color map for the current component/channel.
    x_slice: float, default: None
        The position in world space of the X slicing plane.
    y_slice: float, default: None
        The position in world space of the Y slicing plane.
    z_slice: float, default: None
        The position in world space of the Z slicing plane.
    interpolation: bool, deafult: True
        Linear as opposed to nearest neighbor interpolation for image slices.
        Note: Interpolation is not currently supported with label maps.
    gradient_opacity: float, default: 0.5
        Gradient opacity for composite volume rendering, in the range
        (0.0, 1.0].
    gradient_opacity_scale: float, default: 0.5
        Gradient opacity scale for composite volume rendering, in the range
        (0.0, 1.0].
    gaussians: dict
        Volume rendering opacity transfer function Gaussian parameters. For
        each image component, multiple Gaussians can be specified.
        Default Gaussian parameters:
          {'position': 0.5, 'height': 1, 'width': 0.5, 'xBias': 0.51, 'yBias': 0.4}
    blend_mode: string, default: 'Composite'
        Volume rendering blend mode. Supported modes: 'Composite', 'Maximum',
        'Minimum', 'Average'.
    component_visible: bool, default: True
        The given image intensity component index's visibility.
    shadow_enabled: bool, default: True
        Whether to used gradient-based shadows in the volume rendering.
    view_mode: 'x', 'y', 'z', or 'v', default: 'v'
        Only relevant for 3D scenes.
        Viewing mode:
            'x': x-plane
            'y': y-plane
            'z': z-plane
            'v': volume rendering
    layer: string
        Select the layer identified by `name` in the user interface.
    layer_visible: bool, deafult: True
        Whether the current layer is visible.

    Point Sets
    ^^^^^^^^^^
    point_sets: point set
        The point sets to visualize.

    Other Parameters
    ----------------
    sample_distance: float, default: 0.2
        Sampling distance for volume rendering, normalized from 0.0 to 1.0.
        Lower values result in a higher quality rendering. High values improve
        the framerate.
    units: string, default: ''
        Units to display in the scale bar.

    Returns
    -------
    viewer : ipywidget
        Display by placing at the end of a Jupyter or Colab cell. Query or set
        properties on the object to change the visualization.
    """
    viewer = Viewer(data=data, **kwargs)

    return viewer



"""itkwidgets: Interactive widgets to visualize images, point sets, and 3D geometry on the web."""

__version__ = "1.0a6"

from imjoy_rpc import register_default_codecs
register_default_codecs()

from .imjoy import register_itkwasm_imjoy_codecs
register_itkwasm_imjoy_codecs()

from .viewer import Viewer, view

__all__ = [
  "Viewer",
  "view",
]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
from pathlib import Path
from sphinx.application import Sphinx
import subprocess
import os
import json

# -- Project information -----------------------------------------------------

project = 'itkwidgets'
copyright = '2022, Matthew McCormick'
author = 'Matthew McCormick'

# The full version, including alpha/beta/rc tags
release = '1.0a6'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['myst_parser',]

html_theme_options = dict(
    github_url='https://github.com/InsightSoftwareConsortium/itkwidgets'
)

# jupyterlite_config = jupyterlite_dir / "jupyterlite_config.json"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static',
        'jupyterlite/_output']

def jupyterlite_build(app: Sphinx, error):
    here = Path(__file__).parent.resolve()
    jupyterlite_config = here / "jupyterlite" / "jupyterlite_config.json"
    subprocess.check_call(['jupyter', 'lite', 'build', '--config',
        str(jupyterlite_config)], cwd=str(here / 'jupyterlite'))

def setup(app):
    # For local builds, you can run jupyter lite build manually
    # $ cd jupyterlite
    # $ jupyter lite serve --config ./jupyterlite_config.json
    app.connect("config-inited", jupyterlite_build)



HAVE_DASK = False
try:
    import dask.array
    HAVE_DASK = True
except ImportError:
    pass

import numpy as np


def dask_array_to_ndarray(dask_array):
    array = np.asarray(dask_array)
    array = np.ascontiguousarray(array)
    # JavaScript does not support 64-bit integers
    if array.dtype == np.int64:
        array = array.astype(np.float32)
    elif array.dtype == np.uint64:
        array = array.astype(np.float32)
    return array



def init_params_dict(itk_viewer):
    return {
        'annotations': itk_viewer.setAnnotationsEnabled,
        'axes': itk_viewer.setAxesEnabled,
        'bg_color': itk_viewer.setBackgroundColor,
        'blend_mode': itk_viewer.setImageBlendMode,
        'cmap': itk_viewer.setImageColorMap,
        'color_range': itk_viewer.setImageColorRange,
        'color_bounds': itk_viewer.setImageColorRangeBounds,
        'component_visible': itk_viewer.setImageComponentVisibility,
        'gradient_opacity': itk_viewer.setImageGradientOpacity,
        'gradient_opacity_scale': itk_viewer.setImageGradientOpacityScale,
        'interpolation': itk_viewer.setImageInterpolationEnabled,
        'gaussians': itk_viewer.setImagePiecewiseFunctionGaussians,
        'shadow_enabled': itk_viewer.setImageShadowEnabled,
        'sample_distance': itk_viewer.setImageVolumeSampleDistance,
        'label_blend': itk_viewer.setLabelImageBlend,
        'label_names': itk_viewer.setLabelImageLabelNames,
        'label_lut': itk_viewer.setLabelImageLookupTable,
        'label_weights': itk_viewer.setLabelImageWeights,
        'layer': itk_viewer.selectLayer,
        'layer_visible': itk_viewer.setLayerVisibility,
        'container_style': itk_viewer.setRenderingViewContainerStyle,
        'rotate': itk_viewer.setRotateEnabled,
        'ui_collapsed': itk_viewer.setUICollapsed,
        'units': itk_viewer.setUnits,
        'view_mode': itk_viewer.setViewMode,
        'x_slice': itk_viewer.setXSlice,
        'y_slice': itk_viewer.setYSlice,
        'z_slice': itk_viewer.setZSlice,
    }

def init_key_aliases():
    return {
        'data': 'image',
        'image': 'image',
        'label_image': 'labelImage',
        'point_sets': 'pointSets',
    }



HAVE_XARRAY = False
try:
    import xarray
    HAVE_XARRAY = True
except ImportError:
    pass

def xarray_data_array_to_numpy(data_array):
    return data_array.to_numpy()

def xarray_data_set_to_numpy(data_set):
    return xarray_data_array_to_numpy(data_set.to_array(name='Dataset'))



from dataclasses import dataclass, asdict

from typing import Dict

import itkwasm
import numcodecs
from imjoy_rpc import api

_numcodec_encoder = numcodecs.Blosc(cname='lz4', clevel=3)
_numcodec_config = _numcodec_encoder.get_config()

def encode_itkwasm_image(image):
    global _numcodec_encoder

    image_dict = asdict(image)

    image_data = image_dict['data']
    encoded_data = _numcodec_encoder.encode(image_data)
    image_dict['data'] = { 'buffer': encoded_data, 'config': _numcodec_config, 'nbytes': image_data.nbytes }

    image_direction = image_dict['direction']
    encoded_direction = _numcodec_encoder.encode(image_direction)
    image_dict['direction'] = { 'buffer': encoded_direction, 'config': _numcodec_config, 'nbytes': image_direction.nbytes }

    return image_dict

def register_itkwasm_imjoy_codecs():

    api.registerCodec({'name': 'itkwasm-image', 'type': itkwasm.Image, 'encoder': encode_itkwasm_image})



import itkwasm

from packaging import version
HAVE_ITK = False
try:
    import itk
    if not hasattr(itk, '__version__') or version.parse(itk.__version__) < version.parse('5.3.0'):
      raise RuntimeError('itk 5.3rc4 or newer is required. `pip install --upgrade --pre itk`')
    HAVE_ITK = True
except ImportError:
    pass
    

if HAVE_ITK:
    def itk_image_to_wasm_image(image):
        image_dict = itk.dict_from_image(image)
        wasm_image = itkwasm.Image(**image_dict)
        return wasm_image

    def itk_group_spatial_object_to_wasm_point_set(point_set):
        point_set_dict = itk.dict_from_pointset(point_set)
        wasm_point_set = itkwasm.PointSet(**point_set_dict)
        return wasm_point_set

else:
    def itk_image_to_wasm_image(image):
        raise RuntimeError('itk 5.3rc4 or newer is required. `pip install --upgrade --pre itk`')

    def itk_group_spatial_object_to_wasm_point_set(point_set):
        raise RuntimeError('itk 5.3rc4 or newer is required. `pip install --upgrade --pre itk`')



import itkwasm
import numpy as np
import zarr

from .dask import HAVE_DASK, dask_array_to_ndarray
from .itk import HAVE_ITK, itk_image_to_wasm_image, itk_group_spatial_object_to_wasm_point_set
from .pytorch import HAVE_TORCH
from .vtk import HAVE_VTK, vtk_image_to_ndarray, vtk_polydata_to_vtkjs
from .xarray import HAVE_XARRAY, xarray_data_array_to_numpy, xarray_data_set_to_numpy
from ..render_types import RenderType


async def _get_viewer_image(image):
    if HAVE_ITK:
        import itk
        if isinstance(image, itk.Image):
            return itk_image_to_wasm_image(image)
    if HAVE_VTK:
        import vtk
        if isinstance(image, vtk.vtkImageData):
            return vtk_image_to_ndarray(image)
    if HAVE_DASK:
        import dask
        if isinstance(image, dask.array.core.Array):
            return dask_array_to_ndarray(image)
    if HAVE_TORCH:
        import torch
        if isinstance(image, torch.Tensor):
            return image.numpy()
    if HAVE_XARRAY:
        import xarray
        if isinstance(image, xarray.DataArray):
            return xarray_data_array_to_numpy(image)
        if isinstance(image, xarray.Dataset):
            return xarray_data_set_to_numpy(image)


async def _get_viewer_point_sets(itk_viewer, point_sets):
    if HAVE_VTK:
        import vtk
        if isinstance(point_sets, vtk.vtkPolyData):
            return vtk_polydata_to_vtkjs(point_sets)
    if HAVE_DASK:
        import dask
        if isinstance(point_sets, dask.array.core.Array):
            return dask_array_to_ndarray(point_sets)
    if HAVE_TORCH:
        import torch
        if isinstance(point_sets, torch.Tensor):
            return point_sets.numpy()
    if HAVE_XARRAY:
        import xarray
        if isinstance(point_sets, xarray.DataArray):
            return xarray_data_array_to_numpy(point_sets)
        if isinstance(point_sets, xarray.Dataset):
            return xarray_data_set_to_numpy(point_sets)


def _detect_render_type(data, input_type) -> RenderType:
    if isinstance(data, itkwasm.Image):
        return RenderType.IMAGE
    elif isinstance(data, itkwasm.PointSet):
        return RenderType.POINT_SET
    elif isinstance(data, np.ndarray):
        if input_type == 'point_sets':
            return RenderType.POINT_SET
        else:
            return RenderType.IMAGE
    elif isinstance(data, zarr.Group):
        if input_type == 'point_sets':
            return RenderType.POINT_SET
        else:
            return RenderType.IMAGE
    elif HAVE_ITK:
        import itk
        if isinstance(data, itk.Image):
            return RenderType.IMAGE
    if HAVE_VTK:
        import vtk
        if isinstance(data, vtk.vtkImageData):
            return RenderType.IMAGE
        elif isinstance(data, vtk.vtkPolyData):
            return RenderType.POINT_SET
    if HAVE_DASK:
        import dask
        if isinstance(data, dask.array.core.Array):
            if input_type == 'point_sets':
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
    if HAVE_TORCH:
        import torch
        if isinstance(data, torch.Tensor):
            if input_type == 'point_sets':
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
    if HAVE_XARRAY:
        import xarray
        if isinstance(data, xarray.DataArray):
            if input_type == 'point_sets':
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
        if isinstance(data, xarray.Dataset):
            if input_type == 'point_sets':
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE



import itkwasm
import zarr
import numpy as np

from .integrations.dask import HAVE_DASK
from .integrations.itk import HAVE_ITK
from .integrations.pytorch import HAVE_TORCH
from .integrations.vtk import HAVE_VTK
from .integrations.xarray import HAVE_XARRAY
from typing import Dict, List, Union

Gaussian_Curve = Dict[str, float]
Gaussians = Dict[str, List[Gaussian_Curve]]

Style = Dict[str, str]

Image = Union[np.ndarray, itkwasm.Image, zarr.Group]
Point_Sets = Union[np.ndarray, itkwasm.PointSet, zarr.Group]
if HAVE_ITK:
    import itk
    Image = Union[Image, itk.Image]
if HAVE_VTK:
    import vtk
    Image = Union[Image, vtk.vtkImageData]
    Point_Sets = Union[Point_Sets, vtk.vtkPolyData]
if HAVE_DASK:
    import dask
    Image = Union[Image, dask.array.core.Array]
    Point_Sets = Union[Point_Sets, dask.array.core.Array]
if HAVE_TORCH:
    import torch
    Image = Union[Image, torch.Tensor]
    Point_Sets = Union[Point_Sets, torch.Tensor]
if HAVE_XARRAY:
    import xarray
    Image = Union[Image, xarray.DataArray, xarray.Dataset]
    Point_Sets = Union[Point_Sets, xarray.DataArray, xarray.Dataset]



#!/usr/bin/env python3

import subprocess
import sys

if len(sys.argv) < 2:
    print('Usage: ' + sys.argv[0] + ' <ReleaseNotesFilePath.md>')
    sys.exit(1)
output_file = sys.argv[1]

with open(output_file, 'w') as fp:
    tags = subprocess.check_output(['git', 'tag', '--sort=creatordate']).decode('utf-8')
    recent_tags = tags.split()[-2:]
    previous_tag = recent_tags[0]
    current_tag = recent_tags[1]

    version = current_tag[1:]
    fp.write('# itkwidgets {0}\n\n'.format(version))

    fp.write('`itkwidgets` provides interactive widgets to visualize images, point sets, and 3D geometry on the web.\n\n')

    fp.write('## Installation\n\n')
    fp.write('```\n')
    fp.write('pip install itkwidgets\n')
    fp.write('```\n')
    fp.write('or\n')
    fp.write('```\n')
    fp.write('conda install -c conda-forge itkwidgets\n')
    fp.write('```\n')
    fp.write('\n')

    fp.write('## Changes from {0} to {1}\n'.format(previous_tag, current_tag))
    subjects = subprocess.check_output(['git', 'log', '--pretty=%s:%h',
        '--no-merges', '{0}..{1}'.format(previous_tag, current_tag)]).decode('utf-8')

    bug_fixes = []
    platform_fixes = []
    doc_updates = []
    enhancements = []
    performance_improvements = []
    style_changes = []
    for subject in subjects.split('\n'):
        prefix = subject.split(':')[0]
        commit = subject.split(':')[-1]
        if prefix == 'BUG':
            description = subject.split(':')[1]
            bug_fixes.append((description, commit))
        elif prefix == 'COMP':
            description = subject.split(':')[1]
            platform_fixes.append((description, commit))
        elif prefix == 'DOC':
            description = subject.split(':')[1]
            doc_updates.append((description, commit))
        elif prefix == 'ENH':
            description = subject.split(':')[1]
            enhancements.append((description, commit))
        elif prefix == 'PERF':
            description = subject.split(':')[1]
            performance_improvements.append((description, commit))
        elif prefix == 'STYLE':
            description = subject.split(':')[1]
            style_changes.append((description, commit))

    commit_link_prefix = 'https://github.com/InsightSoftwareConsortium/itkwidgets/commit/'

    if enhancements:
        fp.write('\n### Enhancements\n\n')
        for subject, commit in enhancements:
            if subject.find('Bump itkwidgets version for development') != -1:
                continue
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))

    if performance_improvements:
        fp.write('\n### Performance Improvements\n\n')
        for subject, commit in performance_improvements:
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))

    if doc_updates:
        fp.write('\n### Documentation Updates\n\n')
        for subject, commit in doc_updates:
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))

    if platform_fixes:
        fp.write('\n### Platform Fixes\n\n')
        for subject, commit in platform_fixes:
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))

    if bug_fixes:
        fp.write('\n### Bug Fixes\n\n')
        for subject, commit in bug_fixes:
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))

    if style_changes:
        fp.write('\n### Style Changes\n\n')
        for subject, commit in style_changes:
            fp.write('- {0}'.format(subject))
            fp.write(' ([{0}]({1}{0}))\n'.format(commit, commit_link_prefix))
